@@ -13,7 +13,7 @@ from fabric import Connection, task
 from invoke import Collection  # noqa, re-export
 
 
-__version__ = "1.0.20201223"
+__version__ = "1.0.20210202"
 
 
 # I don't care, in this context.
@@ -29,6 +29,10 @@ red = ansi("31")
 green = ansi("32")
 
 
+def progress(msg):
+    print(green(msg))
+
+
 def terminate(msg):
     print(red(msg), file=sys.stderr)
     sys.exit(1)
@@ -49,6 +53,8 @@ def run(c, *a, **kw):
 
 class Config:
     app = "app"
+    environment = "default"
+    environments = {}
 
     def update(self, **kwargs):
         for key, value in kwargs.items():
@@ -58,9 +64,9 @@ class Config:
         pre_commit_hook()
 
     def __getattr__(self, key):
-        environments = self.__dict__.get("environments", "")
+        environments = getattr(self, "environments", None)
         if environments:
-            environments = f" [{', '.join(sorted(environments))}]"
+            environments = f" [{', '.join(environments)}]"
         terminate(
             f"Configuration key '{key}' not set. "
             f"Maybe you forgot to set an environment with which to interact?"
@@ -72,6 +78,18 @@ class Config:
 config = Config()
 
 
+def environment(name, cfg, **kwargs):
+    config.environments[name] = cfg
+
+    @task(name=name, **kwargs)
+    def fn(ctx):
+        cfg["environment"] = name
+        config.update(**cfg)
+
+    fn.__doc__ = f'Set the environment to "{name}"'
+    return fn
+
+
 class Connection(Connection):
     """Connection subclass which always forwards the agent by default"""
 
@@ -136,7 +154,7 @@ def hook(ctx):
 @task(auto_shortflags=False)
 def dev(ctx, host="127.0.0.1", port=8000):
     """Run the development server for the frontend and backend"""
-    print(green(f"Starting server at http://{host}:{port}/"))
+    progress(f"Starting server at http://{host}:{port}/")
     _concurrently(
         ctx,
         [
@@ -165,7 +183,7 @@ def pull_db(ctx):
 
 @task
 def reset_pw(ctx):
-    """Set all user passwords to ``"password"``"""
+    """Set all user passwords to "password" """
     # 'password' encoded with a constant salt. Does not force a login after pull_db
     pw = r"pbkdf2_sha256\$216000\$salt\$xuFh/Jmp9ZyNeO4k67igyjH9t5hHZ84M69rSfrV2W/g="
     run(
@@ -228,7 +246,7 @@ def _python3():
 
 
 def _pip_up(c):
-    run(c, "venv/bin/python -m pip install -U 'pip!=20.3.2' wheel setuptools")
+    run(c, "venv/bin/python -m pip install -U pip wheel setuptools")
 
 
 @task(auto_shortflags=False, help={"stable": "Avoid pre-release versions of packages"})
@@ -305,15 +323,14 @@ def local(ctx):
 @task
 def nine_vhost(ctx):
     """Create a virtual host using nine-manage-vhosts"""
-    with Connection(config.host) as conn:
+    with Connection(config.host) as conn, conn.cd(config.domain):
         run(
             conn,
             f"sudo nine-manage-vhosts virtual-host create {config.domain}"
             " --template=feinheit_cache"
             f" --webroot=/home/www-data/{config.domain}/htdocs",
         )
-        with conn.cd(config.domain):
-            run(conn, "mkdir -p media tmp")
+        run(conn, "mkdir -p media tmp")
 
 
 @task(auto_shortflags=False, help={"include-www": "Include the www. subdomain"})
@@ -443,13 +460,16 @@ def nine_ssl(ctx):
         )
 
 
+def _nine_restart(conn):
+    run(conn, f"systemctl --user restart gunicorn@{config.domain}.service")
+
+
 @task
 def nine_restart(ctx):
     """Restart the application server"""
-    with Connection(config.host) as conn:
-        with conn.cd(config.domain):
-            run(conn, "venv/bin/python manage.py check --deploy")
-        run(conn, f"systemctl --user restart gunicorn@{config.domain}.service")
+    with Connection(config.host) as conn, conn.cd(config.domain):
+        run(conn, "venv/bin/python manage.py check --deploy")
+        _nine_restart(conn)
 
 
 @task
@@ -484,12 +504,44 @@ def nine_checkout(ctx):
 @task
 def nine_venv(ctx):
     """Create a venv and install packages from requirements.txt"""
+    with Connection(config.host) as conn, conn.cd(config.domain):
+        run(conn, "rm -rf venv")
+        run(conn, "PATH=~/.pyenv/shims:$PATH python3 -m venv venv")
+        _pip_up(conn)
+        run(conn, "venv/bin/python -m pip install -r requirements.txt")
+
+
+@task
+def nine_reinit_from(ctx, environment):
+    """Reinitialize an environment from a different environment"""
+    try:
+        source = config.environments[environment]
+    except KeyError:
+        terminate(f'Unknown source environment "{environment}"')
+
     with Connection(config.host) as conn:
-        with conn.cd(config.domain):
-            run(conn, "rm -rf venv")
-            run(conn, "PATH=~/.pyenv/shims:$PATH python3 -m venv venv")
-            _pip_up(conn)
-            run(conn, "venv/bin/python -m pip install -r requirements.txt")
+        source_e = _srv_env(conn, f"{config.environments[environment]['domain']}/.env")
+        target_e = _srv_env(conn, f"{config.domain}/.env")
+
+        source_dsn = source_e("DATABASE_URL")
+        target_dsn = target_e("DATABASE_URL")
+
+        dbname = _dbname_from_dsn(target_dsn)
+        run(conn, f"pg_dump -Ox {target_dsn} > {config.domain}/tmp/{dbname}.sql")
+        run(conn, f'source ~/.profile && psql -c "DROP DATABASE IF EXISTS {dbname}"')
+        run(
+            conn,
+            f'source ~/.profile && psql -c "CREATE DATABASE {dbname} WITH'
+            f" OWNER {dbname} TEMPLATE template0 ENCODING 'UTF8'"
+            f'"',
+        )
+        run(conn, f"source ~/.profile && pg_dump -Ox {source_dsn} | psql {target_dsn}")
+        run(
+            conn,
+            f"rsync -aH --stats {source['domain']}/media/ {config.domain}/media/",
+        )
+    progress(f"Success! (A database backup is at {config.domain}/tmp/)")
+    progress("You may have to run nine-restart or even deploy once.")
 
 
 @task
@@ -587,12 +639,14 @@ def _check_eslint(ctx):
 
 
 def _check_large_files(ctx, *, limit=500000):
-    out = run(ctx, "git diff-index --cached --name-only HEAD", hide=True).stdout.strip()
+    out = run(
+        ctx, "git diff-index --cached --name-only --diff-filter=AMT HEAD", hide=True
+    ).stdout.strip()
     sizes = ((f, os.stat(f).st_size) for f in out.splitlines())
     large = {f: size for f, size in sizes if size > limit}
     if large:
         files = ", ".join(
-            f"{f} ({size} bytes)"
+            f"{f} ({size // 1000}kb)"
             for f, size in sorted(large.items(), key=lambda r: r[1])
         )
         terminate(f"Large files detected: {files}")
@@ -605,11 +659,10 @@ def _check_branch(ctx):
 
 
 def _check_no_uncommitted_changes(ctx):
-    with Connection(config.host) as conn:
-        with conn.cd(config.domain):
-            result = run(conn, "git status --porcelain").stdout.strip()
-            if result:
-                terminate("Terminating because of uncommitted changes on server")
+    with Connection(config.host) as conn, conn.cd(config.domain):
+        result = run(conn, "git status --porcelain").stdout.strip()
+        if result:
+            terminate("Terminating because of uncommitted changes on server")
 
 
 @task
@@ -655,6 +708,27 @@ def fmt(ctx):
     _fmt_prettier(ctx)
 
 
+def _deploy_django(conn):
+    run(conn, f"git checkout {config.branch}")
+    run(conn, "git fetch origin")
+    run(conn, f"git merge --ff-only origin/{config.branch}")
+    run(conn, 'find . -name "*.pyc" -delete')
+    _pip_up(conn)
+    run(conn, "venv/bin/python -m pip install -r requirements.txt")
+    run(conn, "venv/bin/python manage.py migrate")
+    run(conn, "venv/bin/python manage.py check --deploy", warn=True)
+
+
+def _deploy_staticfiles(conn):
+    run(conn, "venv/bin/python manage.py collectstatic --noinput")
+
+
+def _rsync_static(ctx, *, delete=False):
+    flags = "-pthrz --stats"
+    delete = " --delete" if delete else ""
+    run(ctx, f"rsync {flags}{delete} static/ {config.host}:{config.domain}/static/")
+
+
 @task(auto_shortflags=False, help={"fast": "Skip the Webpack build"})
 def deploy(ctx, fast=False):
     """Deploy once 🔥"""
@@ -665,26 +739,15 @@ def deploy(ctx, fast=False):
     if not fast:
         run(ctx, "NODE_ENV=production npx webpack -p --bail")
 
-    with Connection(config.host) as conn:
-        with conn.cd(config.domain):
-            run(conn, f"git checkout {config.branch}")
-            run(conn, "git fetch origin")
-            run(conn, f"git merge --ff-only origin/{config.branch}")
-            run(conn, 'find . -name "*.pyc" -delete')
-            _pip_up(conn)
-            run(conn, "venv/bin/python -m pip install -r requirements.txt")
-            run(conn, "venv/bin/python manage.py migrate")
-            run(conn, "venv/bin/python manage.py check --deploy", warn=True)
-            if not fast:
-                run(
-                    ctx,
-                    f"rsync -pthrz --delete --stats"
-                    f" static/ {config.host}:{config.domain}/static/",
-                )
-            run(conn, "venv/bin/python manage.py collectstatic --noinput")
-        run(conn, f"systemctl --user restart gunicorn@{config.domain}.service")
+    with Connection(config.host) as conn, conn.cd(config.domain):
+        _deploy_django(conn)
+        if not fast:
+            _rsync_static(ctx, delete=True)
+        _deploy_staticfiles(conn)
+        _nine_restart(conn)
 
     fetch(ctx)
+    progress(f"Successfully deployed the {config.environment} environment.")
 
 
 GENERAL = {
@@ -716,5 +779,6 @@ NINE = {
     nine_disable,
     nine_checkout,
     nine_venv,
+    nine_reinit_from,
     nine,
 }
